"""
autowire.helpers
================

Helpers.

"""
import contextlib

from .base import BaseContext, BaseResource, Implementation
from .utils import RefCounter


class SharedImplementation(Implementation):
    """
    Share implementation in each contexts.

    See :func:`~autowire.decorators.shared` for more details.

    """
    def __init__(self, impl: Implementation):
        super().__init__()
        self.impl = impl
        self.counters = {}

    @contextlib.contextmanager
    def reify(self, resource: BaseResource, context: BaseContext):
        if context not in self.counters:
            context_manager = self.impl.reify(resource, context)
            self.counters[context] = RefCounter(context_manager)
        counter = self.counters[context]
        try:
            with counter as value:
                yield value
        finally:
            if counter.count == 0:
                del self.counters[context]


class GloballySharedImplementation(Implementation):
    """
    Share implementation in providing context.

    See :func:`~autowire.decorators.shared` for more details.

    """
    def __init__(self, impl: Implementation):
        super().__init__()
        self.impl = impl
        self.counter = None

    @contextlib.contextmanager
    def reify(self, resource: BaseResource, context: BaseContext):
        if self.counter is None:
            provider = context.provided_by(resource)  # Find me
            self.counter = RefCounter(self.impl.reify(resource, provider))
        try:
            with self.counter as value:
                yield value
        finally:
            if self.counter.count == 0:
                self.counter = None



"""
autowire.resource.function
==========================

Resource implementation as a function

"""
import functools

from autowire.base import BaseResource, Implementation
from autowire.impl.implementable import Implementable

from .utils import default_implementation


class FunctionResource(BaseResource, Implementable):
    """
    Resource class that wraps a function.

    """

    def __init__(self, func, *, name=None, namespace=None):
        if name is None:
            name = func.__name__
        if namespace is None:
            namespace = func.__module__
        super().__init__(name, namespace)
        self.func = func
        self.default_implementation = default_implementation
        functools.update_wrapper(self, func)

    @property
    def default_implementation(self):
        return self._default_implementation

    @default_implementation.setter
    def default_implementation(self, default_implementation):
        self._default_implementation = default_implementation

    def implement(self, implementation: Implementation):
        """Set implementation of implementable. ::

            @resource.implement
        
        """
        self.default_implementation = implementation
        return implementation

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)



"""
autowire.context.root
=====================

Root context.

"""
from autowire.base import BaseContext, BaseResource

__all__ = ['root_context']


class _RootContext(BaseContext):
    """
    Root context for all contexts

    """
    def get_implementation(self, resource: BaseResource):
        """Get resource implementation from this context."""
        return resource.default_implementation

    def provided_by(self, resource: BaseResource):
        """Find context that provides resource."""
        return self


#: Parent of all contexts
root_context = _RootContext()
del _RootContext



"""
autowire.decorators
===================

Decorators for autowire.

"""
import contextlib

from .helpers import GloballySharedImplementation, SharedImplementation
from .impl.function import FunctionImplementation


def shared(impl: FunctionImplementation=None, *, globally=False):
    """
    Convert implementation to shared resource implementation.

    Shared resource shares context using reference counting.
    When you try to resolve shared resource that already resolve but not teared
    down, It brings previously resolved resouce. ::

        from contextlib import contextmanager

        from autowire import Context, Resource
        from autowire.decorators import shared

        dog = Resource('dog', __name__)
        walk = Resource('walk', __name__)

        @dog.implement
        @shared
        @impl.implementation
        @contextmanager
        def with_dog(resource, context):
            print("Dog is entering")
            try:
                yield "🐶"
            finally:
                print("Dog leaved")

        @walk.implement
        @impl.implementation
        @contextmanager
        def with_walking(resource, context):
            with context.resolve(dog) as dog_value:
                yield "Walking with {}".format(dog_value)

        context = Context()

        with context.resolve(walk) as message:
            print(message)
            with context.resolve(dog) as dog_value:
                print("Feeding {}".format(dog_value))

        # Output:
        # Dog is entering
        # Walking with 🐶
        # Feeding 🐶
        # Dog leaved

    When set globally to `True` the resource will be shared globally.
    "Globally shared" means resource is shared in providing context.

    Globally shared resource will be tied into resource's providing context.
    It means that children's context will not be able to be used
    in this resource. ::

        res1 = Resource('res1', __name__)
        res2 = Resource('res2', __name__)

        @res2.implement
        @shared(globally=True)
        @impl.implementation
        @contextlib.contextmanager
        def with_res1(resource, context):
            with context.resolve(res2) as res1_factory:
                res1 = res1_factory()
                try:
                    yield res1
                finally:
                    res1.teardown()

        context = Context()

        child = Context(context)

        @child.autowired(res2)
        def with_res2():
            yield ...

        # Get ResourceNotProvidedError, even though child context provides
        # res2. Because res1 was shared globally in parent context.
        with child.resolve(res1):
            ...


    """
    if impl is not None:
        return shared(globally=globally)(impl)

    def decorator(impl: FunctionImplementation):
        if globally:
            shared_impl = GloballySharedImplementation(impl)
        else:
            shared_impl = SharedImplementation(impl)

        # Create a function implementation to preserve interface
        @contextlib.contextmanager
        def evaluator(fn, resource, context):
            with shared_impl.reify(resource, context) as value:
                yield value

        return FunctionImplementation(impl.function, evaluator)
    return decorator



"""
context.provider
================

Interface for providing resource's implementations.

"""
import abc

from autowire.base import BaseResource, Implementation
from autowire.impl.implementable import Implementable


class ImplementationConsumer(object, metaclass=abc.ABCMeta):
    """
    Declarative base type for objects that can bind resources with
    implementation.

    """
    @abc.abstractmethod
    def set_implementation(self, resource: BaseResource,
                           implementation: Implementation):
        pass


class Provider(Implementable):
    """
    Type that can provide implementation to specific resource.

    """
    def __init__(self,
                 resource: BaseResource,
                 consumer: ImplementationConsumer):
        super().__init__()
        self.resource = resource
        self.consumer = consumer

    def implement(self, implementation: Implementation):
        """Set implementation of implementable."""
        self.consumer.set_implementation(self.resource, implementation)

    def __call__(self, implementation: Implementation):
        """Set implementation and return. You can use it as a decorator"""
        self.implement(implementation)
        return implementation



"""
autowire.builtins
=================

Builtin resources.

"""
import contextlib

from .base import BaseContext, BaseResource


#
# Types
#


class ContextResource(BaseResource):
    @property
    def default_implementation(self):
        @contextlib.contextmanager
        def impl(resource: BaseResource, context: BaseContext):
            yield context

        return impl

#
# Resources
#


#: Current context resource
context = ContextResource('context', __name__)



"""
impl.implementable
==================

Abstract base class for implementable things.

"""
import abc

from autowire.base import Implementation


class Implementable(object, metaclass=abc.ABCMeta):
    """Some type that can be implemented"""
    @abc.abstractmethod
    def implement(self, implementation: Implementation) -> Implementation:
        """
        Implement this.

        It returns given implementation so that can be used as a decorator.
        """
        pass



"""
autowire.base
=============

Base definitions of autowire.

"""
import abc
import contextlib
import functools

from autowire._compat import abstractproperty


class BaseResource(object, metaclass=abc.ABCMeta):
    """
    Decalarative contextual resource definition.
    """

    def __init__(self, name: str, namespace: str) -> None:
        """
        Create a resource with name.
        Namespace generally be the module's name.

        `name` cannot include any dot(.) characters.

            >>> resource = Resource('name', __name__)

        """
        super().__init__()
        self.name = name
        self.namespace = namespace

    def is_valid_name(self, name: str) -> bool:
        """
        Resource name validation.
        """
        return '.' not in name

    @property
    def canonical_name(self) -> str:
        """
        Canonical name of resource.

        It's <namespace>.<name>

        """
        return self.namespace + '.' + self.name

    @abstractproperty
    def default_implementation(self) -> 'Implementation':
        pass

    def __repr__(self):
        return "{cls}({name!r}, {namespace!r})".format(
            cls=type(self).__name__,
            name=self.name,
            namespace=self.namespace
        )


class BaseContext(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_implementation(self, resource: BaseResource):
        """Get resource implementation from this context."""
        pass

    @abc.abstractmethod
    def provided_by(self, resource: BaseResource) -> 'BaseContext':
        """Find context that provides resource."""
        pass

    def resolve(self, resource: BaseResource):
        """Resolve resource in this context."""
        impl = self.find_resource_impl(resource)
        return impl.reify(resource, self)

    def run(self, resource: BaseResource):
        """
        Resolve a resource just for side-effect.

        It is useful when you want entry-point resource. ::

            @resource.create
            @impl.autowire('foo', foo)
            @impl.autowire('bar', bar)
            @impl.plain
            def main(foo, bar):
                print(foo)
                print(bar)
                ...

            ...

            if __name__ == '__main__':
                context.run(main)

        """
        with self.resolve(resource) as value:
            return value

    def resolve_all(self, resources):
        """Resolve resources in this context."""
        contexts = [self.resolve(resource) for resource in resources]

        @contextlib.contextmanager
        def merged_context(rest: list):
            if rest:
                first, *rest = rest
                with first as value:
                    with merged_context(rest) as rest_values:
                        yield [value] + rest_values
            else:
                yield []
        return merged_context(contexts)

    def find_resource_impl(self, resource: BaseResource) -> 'Implementation':
        """Find resource implementation from this context and its parents."""
        context = self.provided_by(resource)
        return context.get_implementation(resource)

    def partial(self, *positionals, **keywords):
        """
        New function with partial application of the given resources. ::

            @context.partial(foo_resource, bar=bar_resource)
            def func(foo, baz, *, bar=None):
                print(foo)
                print(bar)
                print(baz)

            func('Bar')
            # Output:
            # Foo
            # Bar
            # Baz

        """
        def decorator(fn):
            @functools.wraps(fn)
            def wrapper(*args, **kwargs):
                keyword_items = list(keywords.items())
                keyword_keys = [k for k, _ in keyword_items]
                keyword_resources = [v for _, v in keyword_items]

                with self.resolve_all(positionals) as resolved_args, \
                        self.resolve_all(
                            keyword_resources) as resolved_kwarg_values:
                    resolved_kwargs = {}
                    for k, v in zip(keyword_keys, resolved_kwarg_values):
                        resolved_kwargs[k] = v
                    partial = functools.partial(
                        fn, *resolved_args, **resolved_kwargs)
                    return partial(*args, **kwargs)
            return wrapper
        return decorator


class Implementation(object, metaclass=abc.ABCMeta):
    """Base implementation type"""
    @abc.abstractmethod
    def reify(self, resource: BaseResource, context: BaseContext):
        pass



"""
_compat
=======

Compatability support

"""
import abc
import sys

MAJOR, MINOR = sys.version_info[:2]
PY33 = MAJOR == 3 and MINOR >= 3
PY35 = MAJOR == 3 and MINOR >= 5


# abc.abstractproperty was deprecated since Python 3.3
if PY33:
    def abstractproperty(getter):
        return property(abc.abstractmethod(getter))
else:
    abstractproperty = abc.abstractproperty



"""
impl.function
=============

Function implementation classes

"""
import abc
import contextlib
import functools
import types

from autowire.base import BaseResource, BaseContext, Implementation


# (FunctionType, BaseResource, BaseContext) -> ContextManager
Evaluator = types.FunctionType


class FunctionImplementation(Implementation, metaclass=abc.ABCMeta):
    """
    Implementation with wrapping function and evaluator

    :param function: function to be wrapped.
    :param evaluator: function evaluator for reifying.
    :type evaluator: (Function, BaseResource, BaseContext) -> ContextManager

    """

    def __init__(self, function: types.FunctionType, evaluator: Evaluator):
        super().__init__()
        self.function = function
        self.evaluator = evaluator
        functools.update_wrapper(self, function)

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    @contextlib.contextmanager
    def reify(self, resource: BaseResource, context: BaseContext):
        with self.evaluator(self.function, resource, context) as value:
            yield value



"""
impl
====

Implementers.

"""
import contextlib
import functools

from autowire.base import BaseContext, BaseResource, Implementation
from autowire.utils import as_contextmanager

from .function import FunctionImplementation
from .implementable import Implementable


def implementation(fn):
    """
    Create an implmentation with function that compatible with reify method.

    The implementation can be called same as original function.

    """
    @contextlib.contextmanager
    def evaluator(fn, resource: BaseResource, context: BaseContext):
        with fn(resource, context) as value:
            yield value

    return FunctionImplementation(fn, evaluator)


def contextmanager(fn):
    """
    Shorcut for combination of ``@contextual`` and
    ``@contextlib.contextmanager`` ::

        @impl.contextual
        @contextlib.contextmanager
        def some_impl():
            ...

    is equivalent to ::

        @impl.contextmanager
        def some_impl(resource, context):
            ...

    """
    return contextual(contextlib.contextmanager(fn))


def contextual(fn):
    """
    Create an implementation with function that returns context manager.

    The implementation can be called same as original function.

    """
    @contextlib.contextmanager
    def evaluator(fn, resource: BaseResource, context: BaseContext):
        with fn() as value:
            yield value

    return FunctionImplementation(fn, evaluator)


def plain(fn):
    """
    Create an implementation with plain function.

    The implementation can be called same as original function.

    """
    @as_contextmanager
    def evaluator(fn, resource: BaseResource, context: BaseContext):
        return fn()

    return FunctionImplementation(fn, evaluator)


def autowired(argname_or_required, required=None):
    """
    Resolve a required resource and inject into function implementation
    as keyword argument. ::

        dependency = Resource('foo', __name__)

        @autowired('dependency', dependency)
        @impl.plain
        def create_something(dependency):
            return create(dependency)

    Resource's `name` property will be used by default for keyword name. ::

        dependency = Resource('foo', __name__)

        @autowired(dependency)
        @impl.plain
        def create_something(foo):
            return create(foo)

    """
    if required is None:
        argname = argname_or_required.name
        required = argname_or_required
    else:
        argname = argname_or_required

    def wrapper(func_impl: FunctionImplementation):
        if not isinstance(func_impl, FunctionImplementation):
            raise ValueError(
                "autowired should be applied to FunctionImplementation"
            )

        @contextlib.contextmanager
        def evaluator(fn, resource: BaseResource, context: BaseContext):
            with context.resolve(required) as arg:
                kwargs = {argname: arg}
                new_function = functools.partial(func_impl.function, **kwargs)
                with func_impl.evaluator(
                        new_function, resource, context) as value:
                    yield value
        return FunctionImplementation(func_impl.function, evaluator)
    return wrapper


__all__ = [
    'implement',
    'implementation',
    'contextual',
    'plain',
    'autowired',
    'Implementation',
    'Implementable',
]



"""
autowire.resource
=================

Resource implementations.

"""
from autowire import impl
from autowire.base import Implementation

from .plain import Resource
from .function import FunctionResource


def create(implementation: Implementation=None, *, name=None, namespace=None):
    """
    Create a resource with implementation. ::

        @resource.create
        @impl.implementation
        @contextlib.contextmanager
        def some_resource(resource, context: Context):
            with open('output.log', 'w') as output:
                yield output

        with context.resolve(some_resource) as f:
            f.write('...')

    The default name and namespace will be resolved from decorated function.

    """
    def decorator(implementation):
        resource = FunctionResource(
            implementation, name=name, namespace=namespace
        )
        resource.implement(implementation)
        return resource
    if impl is not None:
        return decorator(implementation)
    return decorator


__all__ = [
    'Resource',
    'FunctionResource',
    'create',
]



"""
autowire.resource.utils
=======================

Resource utilties.

"""
import contextlib

from autowire import impl
from autowire.base import BaseResource, BaseContext
from autowire.exc import ResourceNotProvidedError


@impl.implementation
@contextlib.contextmanager
def default_implementation(resource: BaseResource, context: BaseContext):
    """No such resource."""
    raise ResourceNotProvidedError(
        "No such resource {resource}".format(
            resource=resource.canonical_name,
        ))
    yield None



"""
autowire.resource.plain
=======================

Plain resource implementation

"""
from autowire.base import BaseResource, Implementation
from autowire.impl.implementable import Implementable

from .utils import default_implementation


class Resource(BaseResource, Implementable):
    """
    Resource class of some functionalities like default implementation.

    """
    def __init__(self, name, namespace):
        super().__init__(name, namespace)
        self.default_implementation = default_implementation

    @property
    def default_implementation(self):
        return self._default_implementation

    @default_implementation.setter
    def default_implementation(self, default_implementation):
        self._default_implementation = default_implementation

    def implement(self, implementation: Implementation):
        """Set implementation of implementable."""
        self.default_implementation = implementation



"""
autowire.utils
==============

Common utilities.

"""
import contextlib
import functools


class RefCounter(object):
    """Preserve context until it has any references."""
    def __init__(self, contextmanager):
        super().__init__()
        self.count = 0
        self.contextmanager = contextmanager
        self.value = None

    def increase(self):
        self.count += 1
        if self.count == 1:
            self.value = self.contextmanager.__enter__()

    def decrease(self, exc_type, exc_value, tb):
        if self.count == 1:
            self.value = 0
            self.contextmanager.__exit__(exc_type, exc_value, tb)
        self.count -= 1

    def __enter__(self):
        self.increase()
        return self.value

    def __exit__(self, exc_type, exc_value, tb):
        self.decrease(exc_type, exc_value, tb)


def as_contextmanager(fn):
    """
    Convert to context manager. ::

        @as_contextmanager
        def foo(name):
            return 'Hello, {}'.format(name)

    is equivalent to ::

        @contextlib.contextmanager
        def foo(name):
            yield 'Hello, {}'.format(name)
    """
    @functools.wraps(fn)
    @contextlib.contextmanager
    def wrapper(*args, **kwargs):
        yield fn(*args, **kwargs)
    return wrapper



import sys

import setuptools
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='Autowire',
    packages=setuptools.find_packages(exclude=['tests']),
    version='0.3.1',
    description="Simple dependency injection.",
    author='Geonu Choi',
    author_email='6566gun@gmail.com',
    url='https://github.com/hardtack/autowire',
    license='MIT LICENSE',
    keywords=['dependency-injection'],
    # Cmd
    cmdclass={
        'test': PyTest,
    }
)



"""
autowire.context.impl
=====================

Context's implementations

"""
from autowire.base import BaseContext, BaseResource, Implementation

from .provider import Provider, ImplementationConsumer
from .root import root_context


class Context(BaseContext, ImplementationConsumer):
    """
    Context implementation.
    """

    def __init__(self, parent: BaseContext=root_context):
        self.parent = parent
        self.implementations = {}

    def get_implementation(self, resource: BaseResource):
        """Get resource implementation from this context."""
        return self.implementations[resource.canonical_name]

    def provided_by(self, resource: BaseResource):
        """Find context that provides resource."""
        if resource.canonical_name in self.implementations:
            return self
        elif self.parent is not None:
            return self.parent.provided_by(resource)

    def set_implementation(self, resource: BaseResource,
                           implementation: Implementation):
        self.implementations[resource.canonical_name] = implementation

    def provide(self,
                resource: BaseResource,
                implementation: Implementation=None):
        """
        Provide an implementaion for resource in this context. ::

            context.provide(resource, implementation)

        Or as a decorator ::

            @context.provide(resource)
            @impl.autowired('dependency1', dependency1)
            @impl.autowired('dependency2', dependency2)
            @impl.contextmanager
            def create_resource(dependency1, dependency2):
                yield dependency1.make_resource(dependency2)

        """
        if implementation is not None:
            return self.set_implementation(resource, implementation)
        return Provider(resource, self)


__all__ = ['Context', 'root_context']



from . import impl
from .context import Context, root_context
from .exc import ResourceNotProvidedError
from .resource import Resource


__all__ = [
    'impl',
    'Context',
    'root_context',
    'ResourceNotProvidedError',
    'Resource',
]
